<?php
/*
 * Contains a class for abstraction of a typical unix-style daemon.
 *
 */

require_once("logger.inc.php");

class Daemon
{
	var $pidfile;
	var $syslogIdent;
	var $daemonize;
	var $programName;
	var $shouldExit;

	function __construct($programName, $syslogIdent)
	{
		$this->programName = $programName;
		$this->pidfile = sprintf("/var/run/%s.pid", $programName);
		$this->syslogIdent = $syslogIdent;
		$this->daemonize = false;
		$this->shouldExit = false;
	}

	function pidFileExists()
	{
		return file_exists($this->pidfile);
	}

	function pidRead()
	{
		return intval(trim(file_get_contents($this->pidfile)));
	}

	function pidWrite($pid)
	{
		return file_put_contents($this->pidfile, sprintf("%u\n", $pid));
	}

	function pidFileDelete()
	{
		return unlink($this->pidfile);
	}

	function parseCmdLine()
	{
		global $argv;
		$progname = array_shift($argv);

		while (count($argv) > 0) {
			$p_switch = array_shift($argv);
			if (in_array($p_switch, array("-p", "--pidfile"))) {
				$this->pidfile = array_shift($argv);
			} else if (in_array($p_switch, array("-d", "--daemonize"))) {
				$this->daemonize = true;
			} else if (in_array($p_switch, array("-h", "--help"))) {
				$this->help();
				exit(0);
			} else if (in_array($p_switch, array("-k", "--kill"))) {
				exit($this->killDaemon());
			} else if (in_array($p_switch, array("-v", "--verbose"))) {
				logVerbosityIncrease();
			} else {
				printf("Unknown option: %s\n", $p_switch);
				$this->help();
				exit(1);
			}
		}
	}

	function help()
	{
		echo sprintf("Help for %s\n", $this->programName);
		echo "  -p | --pidfile - indicates the location of pidfile for child\n";
		echo "  -d | --daemonize - indicates that the daemon should be run in\n";
		echo "                     a forked instance\n";
		echo "  -k | --kill - kill of any daemonized instance.\n";
		echo "  -v | --verbose - increase verbosity level.\n";
	}

	function run()
	{
		return $this->main();
	}

	function isDaemonRunning()
	{
		$pid = $this->pidRead();
		if (! $pid) {
			return false;
		}

		return posix_kill($pid, 0);
	}

	function killDaemon()
	{
		if (! $this->pidFileExists()) {
			Logger::log("Pidfile does not exist, exiting...");
			exit(1);
		}
		if (! $this->isDaemonRunning()) {
			Logger::log("Stale pidfile found, removing... ");
			unlink($this->pidfile);
			exit(2);
		}
		$pid = $this->pidRead();
		if (! $pid) {
			Logger::log("Could not read pid from pidfile '%s'", $this->pidfile);
			exit(2);
		}

		Logger::log("Killing daemon with pid %d", $pid);
		posix_kill($pid, SIGTERM);
		sleep(2);
		posix_kill($pid, SIGKILL);
		unlink($this->pidfile);

		exit(0);
	}

	function ensureCommandLine()
	{
		if (!isset($_SERVER["argv"][0])) {
			die("<br><strong>This script is only meant to run at the command line.</strong>");
		}
	}

	function main()
	{
		$this->ensureCommandLine();
		$this->parseCmdLine();
		if (! $this->init())
			exit(0);

		if ($this->daemonize) {
			if ($this->startDaemon())
				exit(0);
			else
				exit(1);
		} else {
			exit($this->daemonMain());
		}
	}

	function init()
	{
		// Install signal handler
		if (! pcntl_signal(SIGINT, array(& $this, "signalHandler"))) {
			Logger::log("Error installing signal handler");
			return false;
		}

		set_error_handler(array(& $this, "phpErrorHandler"));

		// Set default timezone to get rid of lots of php warnings
		date_default_timezone_set(date_default_timezone_get());

		return true;
	}

	function phpErrorHandler($errno, $errstr, $errfile, $errline)
	{
		$msg = sprintf("PHPError File: %s Line: %u Msg: %s", $errfile, $errline, $errstr);
		$level = LOG_DEBUG;
		switch($errno) {
		 case E_USER_ERROR:
		 case E_ERROR:
			$level = LOG_ERR;
			break;
		 case E_USER_WARNING:
		 case E_WARNING:
			$level = LOG_WARNING;
			break;
		 case E_USER_NOTICE:
		 case E_NOTICE:
			$level = LOG_NOTICE;
			break;
		 default:
			$level = LOG_DEBUG;
			break;
		}
		//Logger::debug($msg);
	}

	function signalHandler($signal)
	{
		Logger::log("Received signal " . $signal);
		$this->shouldExit = true;
	}

	function daemonMain()
	{
		throw new Exception("You should override daemonMain()!");
	}

	function startDaemon()
	{
		if ($this->pidFileExists()) {
			if ($this->isDaemonRunning()) {
				Logger::log("Daemon is already running with pid %d, exiting... ", $this->pidRead());
				return false;
			} else {
				Logger::log("Daemon pidfile %s found with pid %d, but no process exists. Removing pidfile.", $this->pidfile, $this->pidRead());
				if (! unlink($this->pidfile))
					Logger::log("Error removing pidfile %s", $this->pidfile);
			}
		}
		$pid = pcntl_fork();
		if ($pid == -1) {
			Logger::log("Error forking daemon child");
			return false;
		} else if ($pid == 0) {
			// Child
			Logger::log("Daemon %s starting up... ", $this->programName);
			posix_setsid();
			$this->daemonMain();
			Logger::log("Daemon %s shutting down... ", $this->programName);
			exit(0);
		} else {
			// Parent
			if (! $this->pidWrite($pid)) {
				// We could not write to the pidfile, so we try to kill the child
				// and exit indicating the error
				Logger::log("Could not write to pidfile! Killing child and bailing out..");
				posix_kill($pid, SIGTERM);
				posix_kill($pid, SIGKILL);
				return false;
			}

			return true;
		}
	}

	function getShouldExit()
	{
		return $this->shouldExit;
	}
}
