<?php

require_once("logger.inc.php");
require_once("/opt/op5sys/etc/config-daemon.conf");

class Protocol
{
	const AUTO_COMMIT = 0;
	const MANUAL_COMMIT = 1;
	const REQUIRE_COMMIT = 2;

	private $state;
	private $retOutput;
	private $retCode;
	private $commandList;
	private $pid;

	function __construct()
	{
		$this->state = Protocol::AUTO_COMMIT;
		$this->pid = posix_getpid();
	}

	/**
	 * used for all commands to determine whether they have resticted symbols
	 */
	private function checkSimple($cmd)
	{
		if (preg_match('/[;&|>]/', $cmd))
			return false;

		return true;
	}

	/**
	 * used only for read and edit commands to determine whether
	 * they have additional resticted symbols
	 */
	private function checkExtended($cmd) // used only for read and edit commands
	{
		$command = explode(" ", rtrim($cmd, "\n"), 3);

		if ((count($command)<3)?true:preg_match('/[.]/', $command[1]))
			return false;

		return true;
	}

	/**
	 * Return command name without params
	 */
	private function getCommand($cmd)
	{
		$command = explode(" ", rtrim($cmd, "\n"), 2);
		return $command[0];
	}

	/**
	 * Execute command
	 */
	private function execCommand($cmd)
	{
		global $CONF;

		unset($this->retOutput);
		$command = explode(" ", rtrim($cmd, "\n"));

		/* If count($command) < 3, we don't have a subcommand and that's not valid */
		if (count($command) < 3) {
			$this->retCode = 125;
			$this->retOutput = array("Invalid command");
			return false;
		}

		$command_path = realpath($CONF["COMMAND_PATH"]);
		/*
		 * Just some basic sanity checking
		 */
		if (strlen($command_path) == 0 || realpath($command_path) == "/" )
		{
			$this->retCode = 127;
			$this->retOutput = array("Invalid COMMAND_PATH");
			return false;
		}

		$exec_cmd = $command_path . "/" . $command[1] . "/" . $command[0];
		/* realpath returns FALSE if no real path was parseable or the result
		 * was not a real (existing) path
		 *
		 * By now, we're also certain that this command is located in a subdirectory
		 * of COMMAND_PATH.
		 */
		if ($exec_cmd !== realpath($exec_cmd)) {
			$this->retCode = 127;
			$this->retOutput = array("Unknown command");
			return false;
		}

		/*
		 * This looks like a valid, "whitelisted" command - let's sanitize
		 * the arguments for the benefit of the (bash) command handlers.
		 * */
		$subcommand = escapeshellarg($command[2]);
		$vargs = array($exec_cmd, $subcommand);
		if (count($command) > 3) {
			foreach (array_slice($command, 3) as $k => $v) {
				$vargs[] = escapeshellarg($v);
			}
		}

		$exec_cmd = implode(" ", $vargs);
		Logger::debug("(%d) command to execute : {%s}", $this->pid, $exec_cmd);
		exec($exec_cmd, $this->retOutput, $this->retCode);
		Logger::debug("(%d) Command return code : {%d}", $this->pid, $this->retCode);
		if ($this->retCode)
		{
			return false;
		}

		return true;
	}

	/**
	 * Execute rollback command
	 */
	private function begin()
	{
		if (!$this->execCommand("session sys init"))
		{
			$this->handleError("begin", implode("\n", $this->retOutput));
			return false;
		}

		$this->state = Protocol::MANUAL_COMMIT;
		return true;
	}

	/*
	 * Execute rollback command
	 */
	private function rollback()
	{
		if (!$this->execCommand("session sys rollback"))
		{
			$this->handleError("rollback", implode("\n", $this->retOutput));
			return false;
		}

		$this->state = Protocol::AUTO_COMMIT;
		return true;
	}

	/**
	 * Execute rollback command
	 */
	private function rollbackAndApply()
	{
		if (!$this->execCommand("session sys rollback-apply"))
		{
			$this->handleError("rollback-apply", implode("\n", $this->retOutput));
			return false;
		}

		$this->state = Protocol::AUTO_COMMIT;
		return true;
	}

	/**
	 * Execute commit command
	 */
	private function commit()
	{
		if (!$this->execCommand("session sys commit"))
		{
			$this->handleError("commit", implode("\n", $this->retOutput));
			return false;
		}

		$this->state = Protocol::AUTO_COMMIT;
		return true;
	}

	/**
	 * Error handler
	 */
	private function handleError($cmd, $retCode, $description)
	{
		$this->state = Protocol::AUTO_COMMIT;
		$cmd_type = strtok($cmd, ' ');
		$cmd_name = strtok(' ');
		Logger::log("(%d) Error : on command {%s} {%s} return code = %d\n=======================\n%s\n=======================\n", $this->pid, $cmd_type, $cmd_name, $retCode, $description);
	}

	/**
	 * Main state handle routine
	 */
	private function autoExecCmd($cmd)
	{
		Logger::debug("(%d) State 1(support only read, edit and begin) : {%s}", $this->pid, rtrim($cmd, "\n"));
		if (!$this->checkSimple($cmd))
		{
			$this->handleError($cmd, 13, "Unsafe command");
			return "13|Unsafe command.\0";
		}

		$command = $this->getCommand($cmd);
		switch ($command)
		{
		 case "read":
			if (!$this->checkExtended($cmd))
			{
				$this->handleError($cmd, 13, "Unsafe command");
				return "13|Unsafe command.\0";
			}
			if (!$this->execCommand($cmd))
			{
				$description = implode("\n", $this->retOutput);
				$this->handleError($cmd, $this->retCode, $description);
				return $this->retCode . "|" . $command . " failed." . $description ."\0" ;
			}
			return "0|" . implode("\n", $this->retOutput) . "\0";

		 case "edit":
			if (!$this->checkExtended($cmd))
			{
				$this->handleError($cmd, 13, "Unsafe command");
				return "13|Unsafe command.\0";
			}
			if (!$this->execCommand($cmd))
			{
				$retCode = $this->retCode;
				$description = implode("\n", $this->retOutput);
				$this->handleError($cmd, $this->retCode, $description);
				$this->rollback();
				return $retCode . "|" . $command . " failed." . $description ."\0" ;
			}
			if (!$this->commit())
			{
				return $this->retCode . "|commit failed.\0";
			}
			return "0|" . implode("\n", $this->retOutput) . "\0";

		 case "begin":
			if ($cmd != "begin\n")
			{
				$this->handleError($cmd, 13, "Unsafe command");
				return "13|Unsafe command.\0";
			}
			if (!$this->begin())
			{
				return $this->retCode . "|begin failed.\0";
			}
			unset($this->commandList);
			return false;

		 default:
			$this->handleError($cmd, 127, "Unknown command");
			return "127|Unknown command.\0";
		}

		return "777|Php error.\0";
	}

	/**
	 * Begin/End state handle routine
	 */
	private function manualExecCmd($cmd)
	{
		Logger::debug("(%d) State 2(support only read, edit and end) : {%s}", $this->pid, rtrim($cmd, "\n"));
		if (!$this->checkSimple($cmd))
		{
			$this->handleError($cmd, 13, "Unsafe command");
			return "13|Unsafe command.\0";
		}

		switch ($this->getCommand($cmd))
		{
		 case "read":
		 case "edit":
			if ($this->checkExtended($cmd))
			{
				$this->commandList []= $cmd;
				return false;
			}
			$this->handleError($cmd, 13, "Unsafe command");
			return "13|Unsafe command.\0";

		 case "end":
			if ($cmd != "end\n")
			{
				$this->handleError($cmd, 13, "Unsafe command");
				return "13|Unsafe command.\0";
			}
			$log = array();
			foreach ($this->commandList as $command)
			{
				if (!$this->execCommand($command))
				{
					$retCode = $this->retCode;
					$description = implode("\n", $this->retOutput);
					$this->handleError($cmd, $this->retCode, $description);
					$this->rollbackAndApply();
					return $retCode . "|" . $command . " failed." . $description ."\0" ;
				}
				$log = array_merge($log, $this->retOutput);
			}
			$this->state = Protocol::REQUIRE_COMMIT;
			return "0|" . implode("\n", $log) . "\0";

		 default:
			$this->handleError($cmd, 127, "Unknown command");
			return "127|Unknown command.\0";
		}

		return "777|Php error.\0";
	}

	/**
	 * Commit/Rollback state handle routine
	 */
	private function requireExecCmd($cmd)
	{
		Logger::debug("(%d) State 3(support only commit and rollback) : {%s}", $this->pid, rtrim($cmd, "\n"));
		if (!$this->checkSimple($cmd))
		{
			$this->handleError($cmd, 13, "Unsafe command");
			return "13|Unsafe command.\0";
		}

		$command = $this->getCommand($cmd);
		switch ($command)
		{
		 case "commit":
		 case "rollback":
			if ($cmd == "commit\n")
			{
				$this->commit();
			}
			else
				if ($cmd == "rollback\n")
				{
					$this->rollbackAndApply();
				}
			else
			{
				$this->handleError($cmd, 13, "Unsafe command");
				return "13|Unsafe command.\0";
			}
			return  $this->retCode . "|" . $command . "\0" ;

		 default:
			$this->handleError($cmd, 127, "Unknown command");
			return "127|Unknown command.\0";
		}

		return "777|Php error.\0";
	}

	function execCmd($cmd)
	{
		switch ($this->state)
		{
		 case Protocol::AUTO_COMMIT:
			return $this->autoExecCmd($cmd);

		 case Protocol::MANUAL_COMMIT:
			return $this->manualExecCmd($cmd);

		 case Protocol::REQUIRE_COMMIT:
			return $this->requireExecCmd($cmd);

		 default:
			$this->handleError($cmd, 777, "Unknow state");
			return false;
		}

		return "777|Php error.\0";
	}

	/**
	 * Execute rollback command
	 */
	function rollbackOnTimeout()
	{
		if ($this->state == Protocol::REQUIRE_COMMIT)
		{
			Logger::log("(%d) rollback on timeout", $this->pid);
			$this->rollbackAndApply();
		}

		return false;
	}
}

?>
